/*
/ HomozygosityHaplotyping program with a new algorithm
/
/ Programming by Koichi Hagiwara 7/24/2009
/
*/

#include "myHeader.h"

char	*annotationFilePath;
char	*exclusionFilePath;
char	*targetFilePath;
char	*outputFilePath;
int		noTgPool[ 2 ];
char	**stFile[ 2 ];

//	sumSNP -- Summarize data.
//	pASNP -- Pointers to each data in sumSNP.
//	SNP -- Subject's genotyping data.
//	pDSNP -- Pointers to each data in SNP.
pSummerizedSNPdata sumSNP;
pSummerizedSNPdata * pASNP;
pSNPdata SNP;
pSNPdata* pDSNP;
//	For human, lastChromosome == 23
int		lastChromosome;
char	stringBuffer[ 10 ][ STRINGBUFFERSIZE ];
// number of valid SNP data.
long	totalAnnotationDataCount;
double	conserved, conservedWOX;
//	chrPosition[ 0 ] -- The start of the chromosome 1.
long	*chrPosition;
double	totalGenetDistance, totalGenetDistanceWOX, totalGenetDistanceWOXWOG;
double	averageMajorAlleleFrequencyWOX, averageMinorAlleleFrequencyWOX;
double	threshold;
//	for nEffectiveSNP for patient m (m=0..noPtPool-1 or noCtPool-1) for chromosome n (n = 1..23 for human) is stored in nEffectiveSNP[ tgPool ][ patient * ( lastChromosome - EXCLUDE_X ) + ( chromosome number(=n) - 1 ) ]
long	*nEffectiveSNP[ 2 ];
//	Regions for analysis.  The values are set for all chromosomes even EXCLUDE_X is defined as 1. 
struct	incReg	includeRegion[ 10000 ];
long	noRegions;
int		enum_patients = 0;
int		enum_controls = 1;
int		enum_significance = 2;
struct	colorBar cb[ 5 ];
double memoryAllocated = 0.0;
//	chrNo is the chromosome number starting from 0 i.e. chromosome 1 has chrNo == 0
int		chrNo;
unsigned int	maxOverlap;
unsigned int	*maxOverlapChr;
FILE	**tempFile;
char	stringTgPool[ 2 ][ 12 ];
char	*targetSt[ 2 ];
char	foldOverlapFileFlagPSHeader;
char	*foldOverlapFileFlagTXT;
long	largeGap;

int main(int argc, char *argv[])
{
	long	i, j, tg;
	char	outfilename[ BUFFERSIZE ], infilename[ BUFFERSIZE ];
	FILE	*inf;
	FILE	*outf;
	char	filename1[ BUFFERSIZE ], filename2[ BUFFERSIZE ], tempStr1[ BUFFERSIZE ];
	char	stringBuffer1[ STRINGBUFFERSIZE ];
	unsigned char	gotValue;
	int		flag;
	
	strcpy( stringTgPool[ enum_patients ], "PatientPool" );
	strcpy( stringTgPool[ enum_controls ], "ControlPool" );

	//	Abort when analysis file is not found.
	if ( ( argc != 2 ) || ( NULL == ( inf = fopen ( argv[ 1 ], "r" ) ) ) ) {
		fprintf( stderr, "ERROR: Analysis file not found\n" );
		fprintf( stderr, "Usage: HHAnalysisProg analysisfile\n" );
		exit( 1 );
	}
	fclose ( inf );

	readAnalysisFile( argv[ 1 ] );
#ifndef	AUTOMATIC
	fprintf ( stderr, "Input RCHH cutoff in cM:  " );
	fgets ( tempStr1, 255, stdin );
	sscanf ( tempStr1, " %lf", &threshold );
	fprintf ( stderr, "Cutoff is %.2f cM\n", threshold );
#else
	threshold = CUTOFFcM;
#endif

	// read anotation file and display statistics
	readAnnotation( annotationFilePath );

	//	Set up the region excluded.
	if ( 0 == ( noRegions = setRegion( exclusionFilePath, targetFilePath, chrPosition[ lastChromosome - EXCLUDE_X ] ) ) ) {
		fprintf( stderr, "ERROR: There are no regions to be analyzed.\n" );
		exit( 1 );
	}
	fprintf ( stderr, "Number of regions %ld\n", noRegions );
	// read patient and/or control pool.
	for( tg = 0; tg < 2; ++tg ) {
		if( noTgPool[ tg ] != 0 ) {
			readGenotypingData ( tg, noTgPool[ tg ] );
		}
	}

	// Output the summarized data.
	for( tg = 0; tg < 2; ++tg ) {
		if( noTgPool[ tg ] != 0 ) {
			sprintf( outfilename, "%sHH_HomozygosityHaplotype_Summary_%s.txt", outputFilePath, stringTgPool[ tg ] );
			if ( ( outf = fopen( outfilename, "w" ) ) == NULL ) { fprintf( stderr, "ERROR: Can't output\"%sHH_HomozygosityHaplotype_Summary_%s.txt\"\n", outputFilePath, stringTgPool[ tg ] ); exit( 1 ); } 
			fprintf( outf, "chr\tpos\tgen map\trsID\t'A' frequency\t" );
			for ( j = 0; j < noTgPool[ tg ]; ++j ) {
				fprintf( outf, "%s\t", fileNameBody( stFile[ tg ][ j ] ) );
			}
			fprintf( outf, "No.\n\n" );
			for ( i = 0; i < totalAnnotationDataCount; ++i ) {	
				fprintf( outf, "%d\t%ld\t%f\t%s\t%f\t", sumSNP[ i ].chromosome, sumSNP[ i ].position, sumSNP[ i ].genetMap, sumSNP[ i ].rsID, sumSNP[ i ].AFrequency );
				for ( j = 0; j < noTgPool[ tg ]; ++j ) {
					gotValue = getGType( sumSNP[ i ].gType[ tg ], j );
					if ( gotValue == 1 ) fprintf( outf, "A\t" );
					else if ( gotValue == 2 ) fprintf( outf, "B\t" );
					else fprintf( outf, "\t" );
				}
				fprintf( outf, "%ld\n", i );
			}
			fclose( outf );
		}
	}
	// Output the summarized data---end.

	// Preparation for HH analysis
	if ( ( FILE ** ) NULL == ( tempFile = ( FILE ** ) calloc ( lastChromosome - EXCLUDE_X, sizeof( FILE * ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	if ( ( unsigned int * ) NULL == ( maxOverlapChr = ( unsigned int * ) calloc ( lastChromosome - EXCLUDE_X, sizeof( unsigned int ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	for( tg = 0; tg < 2; ++tg ) {
		if ( NULL == ( targetSt[ tg ] = ( char * )malloc( noTgPool[ tg ] ) ) ) { fprintf( stderr, "ERROR: Cannot allocate memory.\n" ); exit( 1 ); }
	}
	// Preparation for HH analysis...end.
	
	// Determine the homozygosity haplotype for each patient and obtain the fold overlaps.
	for( tg = 0; tg < 2; ++tg ) {
		if( noTgPool[ tg ] != 0 ){
			// Preparation for the analysis
			for( i = 0; i < noTgPool[ tg ]; ++ i ) { targetSt[ tg ][ i ] = 1; }
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				sprintf( outfilename, "%stempFile%03ld.txt", outputFilePath, i );
				if( NULL == ( tempFile[ i ] = fopen( outfilename, "w" ) ) ) { fprintf( stderr, "ERROR: Cannot open temporary file.\n" ); exit( 1 ); }
			}
			if( ( char * ) NULL == ( foldOverlapFileFlagTXT = ( char * ) calloc( noTgPool[ tg ], sizeof( char ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
			fprintf( stderr, "\n\n%s\n", stringTgPool[ tg ] );
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				maxOverlapChr[ i ] = 0;
			}
			// Preparation for the analysis...end
			
#pragma omp parallel  
		{    
#pragma omp for schedule(dynamic, 1) nowait
			for( chrNo = 0; chrNo < lastChromosome - EXCLUDE_X; ++chrNo ) {
				long i2, j2;
				if( chrPosition[ chrNo + 1 ] - chrPosition[ chrNo ] == 0 ) {
					continue;
				}
				for( i2 = chrPosition[ chrNo ]; i2 < chrPosition[ chrNo + 1 ]; ++i2 ) {
					if ( ( unsigned short int * ) NULL == ( ( sumSNP + i2 )->foldOverlap = ( unsigned short int * ) calloc ( noTgPool[ tg ] + 1, sizeof ( unsigned short int ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n"); exit ( 1 ); }
					if ( ( unsigned char * ) NULL == ( sumSNP[ i2 ].expandedGType = ( unsigned char * )calloc( noTgPool[ tg ] + 1, sizeof( unsigned char ) ) ) ) { fprintf( stderr, "ERROR: Cannot allocate memory.\n" ); exit( 1 ); }
					for( j2 = 0; j2 < noTgPool[ tg ]; ++j2 ) {
						sumSNP[ i2 ].expandedGType[ j2 ] = getGType( sumSNP[ i2 ].gType[ tg ], j2 );
					}
				}
				
				//	1..AllSubjects are involved
				countRCHHSharedByMultiplePatients( tg, noTgPool[ tg ], chrNo, 1 );
				for ( j2 = noTgPool[ tg ]; j2 > 1; --j2 ) {
					// cutoff = j; mode = 1;
					if( 1 == Set_up_start_end( j2, 1, tg, chrNo ) ){
						sprintf( outfilename, "%sHH_%03ld_FoldOverlaps_%s_chr%03d.ps", outputFilePath, j2, stringTgPool[ tg ], chrNo + 1 );
						foldOverlapFileFlagPSHeader = 0;
						Output_black_and_whitePS( outfilename, chrNo, j2 );
						sprintf( outfilename, "%sHH_%03ld_FoldOverlaps_%s_chr%03d.txt", outputFilePath, j2, stringTgPool[ tg ], chrNo + 1 );
						Output_foldoverlap_textHH( outfilename, noTgPool[ tg ], j2, tg, chrNo, j2 );
					}
				}
				// Obtain and write representative haplotype
				// foldOverlap is needed to obtain obtainRepresentativeHH
				if( noTgPool[ enum_patients ] != 0 && noTgPool[ enum_controls ] != 0 && tg == enum_patients ) {
					// This routine should be placed here because it requires foldOverlap[]
					obtainRepresentativeHH( chrNo );
					// copy representative haplotype
					for( i2 = chrPosition[ chrNo ]; i2 < chrPosition[ chrNo + 1 ]; ++i2 ) {
						putGType( sumSNP[ i2 ].gType[ 1 ], noTgPool[ 1 ], getGType( sumSNP[ i2 ].gType[ 0 ], noTgPool[ 0 ] ) );
					}
					nEffectiveSNP[ 1 ][ noTgPool[ 1 ] * ( lastChromosome - EXCLUDE_X ) + chrNo ] = nEffectiveSNP[ 0 ][ noTgPool[ 0 ] * ( lastChromosome - EXCLUDE_X ) + chrNo ];
				}
				// Obtain and write representative haplotype...ends.
				for ( i2 = chrPosition[ chrNo ]; i2 < chrPosition[ chrNo + 1 ]; ++i2 ) {
					free( sumSNP[ i2 ].foldOverlap );
					free( sumSNP[ i2 ].expandedGType );
				}
			}
		}
		// omp ends.
			
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				if( maxOverlap < maxOverlapChr[ i ] ) {
					maxOverlap = maxOverlapChr[ i ];
				}
			}
			free( foldOverlapFileFlagTXT );
			fprintf ( stderr, "\nCompleted.\n\n" );
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				fclose( tempFile[ i ] );
			}
			displayHHInformation( noTgPool[ tg ], 0 );
			foldOverlapDrawFigures( tg, noTgPool[ tg ], 0 );
			sprintf ( outfilename, "%sHH_FoldOverlap_Summary_%s.txt", outputFilePath, stringTgPool[ tg ] );
			if ( ( outf = fopen( outfilename, "w" ) ) == NULL ) { fprintf( stderr, "ERROR: Can't output %s\n", outfilename ); exit( 1 ); } 
			fprintf( stderr, "\nWriting %s\n", outfilename );
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				sprintf ( infilename, "%sHH_FoldOverlap_Summary_%s_chr%03ld.txt", outputFilePath, stringTgPool[ tg ], i + 1 );
				inf = fopen( infilename, "r" );
				while( ( char * )NULL != MyFgets( stringBuffer1, STRINGBUFFERSIZE, inf ) ) {
					fprintf( outf, "%s\n", stringBuffer1 );
				}
				fclose( inf );
				remove( infilename );
			}
			fclose( outf );
			
			for ( j = noTgPool[ tg ]; j > 1; --j ) {
				for( flag = 0, i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
					if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
						continue;
					}
					sprintf( infilename, "%sHH_%03ld_FoldOverlaps_%s_chr%03ld.txt", outputFilePath, j, stringTgPool[ tg ], i + 1 );
					if( NULL != ( inf = fopen( infilename, "r" ) ) ) {
						if( flag == 0 ) {
							sprintf( outfilename, "%sHH_%03ld_FoldOverlaps_%s.txt", outputFilePath, j, stringTgPool[ tg ] );
							if ( ( outf = fopen( outfilename, "w" ) ) == NULL ) { fprintf( stderr, "ERROR: Can't output %s\n", outfilename ); exit( 1 ); }
							fprintf( stderr, "Writing %s\n", outfilename );
							flag = 1;
						}
						while( ( char * )NULL != MyFgets( stringBuffer1, STRINGBUFFERSIZE, inf ) ) {
							fprintf( outf, "%s\n", stringBuffer1 );
						}
						fclose( inf );
						remove( infilename );
					}
				}
				if( flag == 1 ) {
					fclose( outf );
				}
			}

			for ( j = noTgPool[ tg ]; j > 1; --j ) {
				for( flag = 0, i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
					if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
						continue;
					}
					sprintf( infilename, "%sHH_%03ld_FoldOverlaps_%s_chr%03ld.ps", outputFilePath, j, stringTgPool[ tg ], i + 1 );
					if( NULL != ( inf = fopen( infilename, "r" ) ) ) {
						if( flag == 0 ) {
							sprintf( outfilename, "%sHH_%03ld_FoldOverlaps_%s.ps", outputFilePath, j, stringTgPool[ tg ] );
							if ( ( outf = fopen( outfilename, "w" ) ) == NULL ) { fprintf( stderr, "ERROR: Can't output %s\n", outfilename ); exit( 1 ); }
							fprintf( stderr, "Writing %s\n", outfilename );
							fprintf( outf, "%%!PS-Adobe-3.1 EPSF-3.0\n" );
							fprintf( outf, "0 setlinewidth\n" );
							fprintf( outf, "1 setgray\n" );
							fprintf( outf, "0.02 18 scale	0 1 moveto 24572.955 0 rlineto 0 22.5 rlineto -24572.955 0 rlineto 0 1 lineto closepath stroke\n" );
							fprintf( outf, "0 setgray\n" );
							flag = 1;
						}
						while( ( char * )NULL != MyFgets( stringBuffer1, STRINGBUFFERSIZE, inf ) ) {
							fprintf( outf, "%s\n", stringBuffer1 );
						}
						fclose( inf );
						remove( infilename );
					}
				}
				if( flag == 1 ) {
					fclose( outf );
				}
			}
		}
		// if( noTgPool[ tg ] != 0 ) end.
		
		// Output representative haplotype
		if( noTgPool[ enum_patients ] != 0 && noTgPool[ enum_controls ] != 0 ) {
			sprintf ( outfilename, "%sHH_Representative_PatientPool.txt", outputFilePath );
			if ( ( outf = fopen( outfilename, "w" ) ) == NULL ) { fprintf( stderr, "ERROR: Can't output %sHH_Representative_PatientPool.txt\n", outputFilePath ); exit( 1 ); } 
			fprintf( stderr, "Writing %s\n", outfilename );
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++ i ) {
				for ( j = chrPosition[ i ]; j < chrPosition[ i + 1 ]; ++j ) {
					fprintf( outf, "%d\t%ld\t", sumSNP[ j ].chromosome, sumSNP[ j ].position );
					gotValue = getGType( sumSNP[ j ].gType[ enum_patients ], noTgPool[ enum_patients ] );
					if ( gotValue == 1 ) fprintf( outf, "AA\t" );
					else if ( gotValue == 2 ) fprintf( outf, "BB\t" );
					else fprintf( outf, "AB\t" );
					fprintf( outf, "%f\t%s\t%ld\n", sumSNP[ j ].AFrequency, sumSNP[ j ].rsID, j );
				}
			}
			fclose( outf );		
		}
		// Output representative haplotype...ends.
	}
	// Determine the homozygosity haplotype for each patient and obtain the fold overlaps...ends.
	
	// Obtain P-values.
	if ( noTgPool[ enum_patients ] != 0 && noTgPool[ enum_controls ] != 0 ){
		sprintf( filename1, "%sHH_P_Value_PatientPool_vs_ControlPool.ps", outputFilePath );
		sprintf( filename2, "%sHH_P_Value_PatientPool_vs_ControlPool.txt", outputFilePath );
		significance ( filename1, filename2 );
	}
	// Obtain P-values---end.

	//	Representative Haplotype
	if ( noTgPool[ enum_patients ] != 0 && noTgPool[ enum_controls ] != 0 ){
		for( tg = 0; tg < 2; ++tg ) {
			// Preparation for the analysis for vs representative
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				sprintf( filename1, "%stempFile%03ld.txt", outputFilePath, i );
				if( NULL == ( tempFile[ i ] = fopen( filename1, "w" ) ) ) { fprintf( stderr, "ERROR: Cannot open temporary file.\n" ); exit( 1 ); }
				maxOverlapChr[ i ] = 0;
			}
			fprintf( stderr, "\nRepresentative: %s\n", stringTgPool[ tg ] );
			// Preparation for the analysis of for vs representative...end
			
#pragma omp parallel  
		{    
#pragma omp for schedule(dynamic, 1) nowait
			for( chrNo = 0; chrNo < lastChromosome - EXCLUDE_X; ++chrNo ) {
				if( chrPosition[ chrNo + 1 ] - chrPosition[ chrNo ] == 0 ) {
					continue;
				}
				long i2, j2;
				for( i2 = chrPosition[ chrNo ]; i2 < chrPosition[ chrNo + 1 ]; ++i2 ) {
					if ( ( unsigned short int * ) NULL == ( ( sumSNP + i2 )->foldOverlap = ( unsigned short int * ) calloc ( noTgPool[ tg ] + 1, sizeof ( unsigned short int ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
					if ( ( unsigned char * ) NULL == ( sumSNP[ i2 ].expandedGType = ( unsigned char * )calloc( noTgPool[ tg ] + 1, sizeof( unsigned char ) ) ) ) { fprintf( stderr, "ERROR: Cannot allocate memory.\n" ); exit( 1 ); }
					for( j2 = 0; j2 < noTgPool[ tg ] + 1; ++j2 ) {
						sumSNP[ i2 ].expandedGType[ j2 ] = getGType( sumSNP[ i2 ].gType[ tg ], j2 );
					}
				}
				countRCHHSharedWithRepresentativePatients( tg, noTgPool[ tg ], chrNo );
				for ( i2 = chrPosition[ chrNo ]; i2 < chrPosition[ chrNo + 1 ]; ++i2 ) {
					free( sumSNP[ i2 ].foldOverlap );
					free( sumSNP[ i2 ].expandedGType );
				}
			}
		}
		// omp ends.
			for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
				if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
					continue;
				}
				fclose( tempFile[ i ] );
				if( maxOverlap < maxOverlapChr[ i ] ) {
					maxOverlap = maxOverlapChr[ i ];
				}
			}
			fprintf ( stderr, "\nCompleted.\n\n" );
			displayHHInformation( noTgPool[ tg ], 1 );
			foldOverlapDrawFigures( tg, noTgPool[ tg ], 1 );
			sprintf( filename1, "%sHH_P_Value_Using_Representative_PatientPool_vs_ControlPool.ps", outputFilePath );
			sprintf( filename2, "%sHH_P_Value_Using_Representative_PatientPool_vs_ControlPool.txt", outputFilePath );
			significance( filename1, filename2 );
		}
	}
	//	Representative Haplotype..ends.

	if ( noTgPool[ enum_patients ] != 0 && noTgPool[ enum_controls ] == 0 ){
		while ( 0 != customCombinationHH() );
	}

	for( i = 0; i < lastChromosome - EXCLUDE_X; ++i ) {
		if( chrPosition[ i + 1 ] - chrPosition[ i ] == 0 ) {
			continue;
		}
		sprintf( filename1, "%stempFile%03ld.txt", outputFilePath, i );
		remove( filename1 );
	}
	free( targetSt[ 0 ] );
	free( targetSt[ 1 ] );
	fprintf( stderr, "\nCompleted!\n" );
	
	return 0;
}
