#ifndef	_MYHEADER_H_

#define	_MYHEADER_H_
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

// Custom parameters
#define	OVERLAPCUTOFF	2			//	0...no pair was found.	1...1 pair was found.....
#define	EXCLUDE_X	1				/*	1 if X chromosome is excluded. 0 if not excluded. */
#define	LARGEGAP	300000			//	A region stops if this exists.  Default value is 3000000.  To use this, define CENTROMERE.
#define SHORT						//	If defined, gives short figures
#define	JACKUP 0					//	For GUI JACKUP = 0.1 to fit into the background.
//#define	AUTOMATIC
//#define	CUTOFFcM 3.0

//	Internal use.
//	Suppress Windows-specific warnings. 
#pragma warning( disable : 4996 )
#define CR 13		
#define LF 10
#define	BUFFERSIZE	1000
#define STRINGBUFFERSIZE 10000		//buffer size used for MyFgets
#define	MAX_NUMBER_OF_PARTS 100

struct summerizedSNPdata {
	int		chromosome;						//	chromosome number starts from 1.
	long	position;
	double	genetMap;
	char	*rsID;
	float	AFrequency;
	long	recordNo;
	unsigned short int	*foldOverlap;		//	to store number of overlaps
	unsigned char	*gType[ 2 ];
	unsigned char	*expandedGType;
	unsigned long	sum[ 3 ];				//	numbers of overlaps		sum[ 2 ] is for significance.  Needs to be long to discriminate more than 256 subjects.
	unsigned short	start;							//	The number of patients who share a common fragment.
	unsigned short	end;
	float	color;
	int		ptInvolved[ 2 ];				//	Storing the number of patients who share a common fragment. The value 1 means that 2 patients are involved.
	float	p;								//	A p value.
};
typedef struct summerizedSNPdata*	pSummerizedSNPdata;

// structs
// the storage for the patient's data
struct SNPdata {
	char	gType;		// SNP types are coded as AA=1 BB=2 AB=3 NoCall and others=4
	char	*rsID;
	long	recordNo;
};
typedef struct SNPdata*	pSNPdata;

struct incReg {
	long	start;
	long	end;
};

struct	colorBar {
	double	color;
	double	number;
};

//	common
int		compareAnnotRsID( const void *a, const void *b );
int		comparePtRsID( const void *a, const void *b );
int		customCombination();
char	*fileNameBody( char	*str );
char	*MyFgets( char * str, int size, FILE * stream );
void	readAnnotation( char *pathToAnnotationFile );
void	readGenotypingData (int tgPool, int	noSubjects );
int		Set_up_start_end( int cutoff, int mode, int tgPool, int chrNo );
void	significance( char *outputfilePS, char *outputfileText );
void	fOutput_start_end_text( char *filename, char *message, int tgPool );

//	HH specific
void	conservedHaplotypeFor2( int pt1, int pt2, int tgPool, int isRepresentative, int chrNo );
int		isHetero ( long i, int pt1, int pt2 );
void	Output_foldoverlap_textHH ( char *filename, int noSubjects, long cutoff, int tgPool, int chrNo, int foldOverlap );
int		customCombinationHH();
void	obtainRepresentativeHH( int i );

void	readAnalysisFile ( char *filename );
double	falsePositive ( double thres, long nPt1, long nPt2, double *mean, int chrNo );
long	setRegion ( char *exclFileName, char *tgtFileName, long annotationNo );
int		isInRegion(	long testNo, long noRegions );
void	Output_black_and_whitePS ( char *filename, int chrNo, int foldOverlap );
double	p_value ( int sampleNo1, int sampleNo2, int positiveNo1, int positiveNo2 );
void	Output_birds_eye_viewPS( char *filename, int tgPool );
void	Output_start_end_text( char *filename, char *message, int tgPool );
void	countRCHHSharedByMultiplePatients( int tgPool, int noSubjects, int chrNo, int areAllSubjectsInvolved );
void	countRCHHSharedWithRepresentativePatients( int tgPool, int noSubjects, int chrNo );
int		main(int argc, char *argv[]);
int split( char * buffer, char ** parts, char *separator, int maxNoParts );
char *strip( char *str );
unsigned char getGType ( unsigned char *startGType, long index );
void putGType ( unsigned char *startGType, long index, unsigned char value );
void displayHHInformation( int noSt, int isRepresentative );
void	foldOverlapDrawFigures( int tgPool, int noSubjects, int isRepresentative );

void	init_genrand( unsigned long s );
void init_by_array( unsigned long init_key[], int key_length );
unsigned long genrand_int32( void );
long genrand_int31( void );
double genrand_real1( void );
double genrand_real2( void );
double genrand_real3( void );
double genrand_res53( void );

#endif
