//
//	obtainRCHHSharedByMultiplePatients.c
//
//	Programming:	Koich Hagiwara
//

#include "myHeader.h"

extern	pSummerizedSNPdata sumSNP;
extern	struct	incReg	includeRegion[];	//	The values are set for all chromosomes even EXCLUDE_X is defined as 1. 
extern	long	noRegions;
extern	char	*outputFilePath;
extern	long	*chrPosition;
extern	char	**stFile[];
extern	int		lastChromosome;
extern	unsigned int	*maxOverlapChr;
extern	unsigned int	maxOverlap;
extern	char	stringTgPool[ 2 ][ 12 ];
extern	char	*targetSt[ 2 ];

void	countRCHHSharedByMultiplePatients( int tgPool, int noSubjects, int chrNo, int areAllSubjectsInvolved )
{	
	long	i, j, k, l, i2, numberOfCombinations, numberOfTargetSt;
	double	numberOfSubjectsPerAsterisk;
	FILE	*outf;
	char	outfilename[ BUFFERSIZE ];

	fflush( stdout );
	for( i = chrPosition[ chrNo ]; i < chrPosition[ chrNo + 1 ]; ++ i ) {
		for( j = 0; j < noSubjects; ++j ) {
			sumSNP[ i ].foldOverlap[ j ] = 0;
		}
	}
	if( areAllSubjectsInvolved == 1 ) {
		numberOfTargetSt = noSubjects;
	} else {
		for( numberOfTargetSt = 0, k = 0; k < noSubjects; ++ k ) {
			if( targetSt[ tgPool ][ k ] == 1 ) {
				numberOfTargetSt++;
			}
		}
	}
	numberOfCombinations = numberOfTargetSt * ( numberOfTargetSt - 1 ) / 2;
	numberOfSubjectsPerAsterisk = numberOfCombinations / 10.0 - 0.001;
	for ( k = 0, l = 0, i = 0 ; i < noSubjects; ++i ) {
		for ( j = i + 1 ; j < noSubjects; ++j ) {
			//	0 means that the analysis is not for the representative.
			if ( ( targetSt[ tgPool ][ i ] == 1 ) && ( targetSt[ tgPool ][ j ] == 1 ) ) {
				conservedHaplotypeFor2( i , j , tgPool, 0, chrNo );
				k++;
			}
			while( k / numberOfSubjectsPerAsterisk > l + 1 ) {
				fprintf( stderr, "\nchromosome %d: %ld/10 completed.", chrNo + 1, l + 1 );
				l++;
			}
		}
	}
	//	sumSNP[].sum is cleared for all SNP sites.
	for ( i = chrPosition[ chrNo ]; i < chrPosition[ chrNo + 1 ]; ++i ) {
		sumSNP[ i ].ptInvolved[ tgPool ] = sumSNP[ i ].sum[ tgPool ] = 0;
	}
	for( i2 = 0; i2 < noRegions; ++ i2 ) {
		if( ( chrPosition[ chrNo ] <= includeRegion[ i2 ].start ) && ( includeRegion[ i2 ].end <= chrPosition[ chrNo + 1 ] - 1 ) ) {
			for ( i = includeRegion[ i2 ].start; i <= includeRegion[ i2 ].end; ++i ) {		//	j moves within a region.
				for ( j = 0; j <= noSubjects; j++ ) {										//	j should be till noOfPatients.
					for ( k = 0, l = 0; k < noSubjects; ++k ) {
						if ( sumSNP[ i ].foldOverlap[ k ] >= j ) ++l;				//	Screen patients whose fold overlap value is more than j + 1.
					}
					if ( l >= j + 1 ) continue;												//	If so, try a larger j.  When j == 0, every time this criteria is fulfilled.
					else {																	//	When no pairs were found, sumSNP[].sum is set to 1.
					sumSNP[ i ].ptInvolved[ tgPool ] = sumSNP[ i ].sum[ tgPool ] = j;	//	Otherwise j is what you wanted, i.e. if j = 3, 3 patients share an RCHH.
						break;																//	Exit from from ( j ).
					}
				}
				// sumSNP[ i ][].sum harbors the number of patients who share a common fragment.
				if ( maxOverlapChr[ chrNo ] < sumSNP[ i ].sum[ tgPool ] ) maxOverlapChr[ chrNo ] = sumSNP[ i ].sum[ tgPool ];
			}
		}
	}
	
	//	Output the FoldOverlap summary data.
	//	Fold overlaps takes the value of 0, 1, 2, 3, 4, ..... , because Fold overlaps indicate the number of patients (excluding him/herself) with whom the patient shares an RCHH.
	if( areAllSubjectsInvolved == 1 ) {
		sprintf ( outfilename, "%sHH_FoldOverlap_Summary_%s_chr%03d.txt", outputFilePath, stringTgPool[ tgPool ], chrNo + 1 );
		if ( ( outf = fopen( outfilename, "w" ) ) == NULL ) {
			fprintf( stderr, "ERROR: Can't output\"%sHH_FoldOverlap_Summary_%s_chr%03d.txt\n", outputFilePath, stringTgPool[ tgPool ], chrNo + 1 );
			exit( 1 );
		}
		if( chrNo == 0 ) {
			fprintf( outf, "chr\tpos\tgen map\trsID\t'A' frequency\t" );
			for ( j = 0; j < noSubjects; ++j ) {
				fprintf( outf, "%s\t", fileNameBody( stFile[ tgPool ][ j ] ) );
			}
			fprintf( outf, "overlaps\tNo.\n" );
		}
		for ( i = chrPosition[ chrNo ]; i < chrPosition[ chrNo + 1 ]; ++i ) {
			fprintf( outf, "%d\t%ld\t%f\t%s\t%f\t", sumSNP[ i ].chromosome, sumSNP[ i ].position, sumSNP[ i ].genetMap, sumSNP[ i ].rsID, sumSNP[ i ].AFrequency );
			if ( isInRegion( i, noRegions ) == 1 ) {
				for ( j = 0; j < noSubjects; ++j ) {
					fprintf( outf, "%d\t", sumSNP[ i ].foldOverlap[ j ] );	//	Note that the max of foldOverlaps is equal to maxOverlap - 1.
				}
				fprintf( outf, "%ld\t%ld\n",sumSNP[ i ].sum[ tgPool ], i );
			} else {
				for ( j = 0; j < noSubjects; ++j ) {
					fprintf( outf, "(%d)\t", sumSNP[ i ].foldOverlap[ j ] );	//	Note that the max of foldOverlaps is equal to maxOverlap - 1.
				}
				fprintf( outf, "(%ld)\t%ld\n", sumSNP[ i ].sum[ tgPool ], i );
			}
		}
		fclose(outf);
	}
	// Output the summarized data---end.
}

void	countRCHHSharedWithRepresentativePatients( int tgPool, int noSubjects, int chrNo )
{	
	long	i, k, l, i2, numberOfCombinations;
	double	numberOfSubjectsPerAsterisk;

	fflush( stdout );
	for( i = chrPosition[ chrNo ]; i < chrPosition[ chrNo + 1 ]; ++ i ) {
		sumSNP[ i ].foldOverlap[ noSubjects ] = 0;
	}
	numberOfCombinations = noSubjects;
	numberOfSubjectsPerAsterisk = numberOfCombinations / 10.0 - 0.001;
	for ( k = 0, l = 0, i = 0 ; i < noSubjects; ++i ) {
		//	1 means that the analysis is for the representative.
		conservedHaplotypeFor2( i , noSubjects, tgPool, 1, chrNo );
		k++;
		while( k / numberOfSubjectsPerAsterisk > l + 1 ) {
			fprintf( stderr, "\nchromosome %d: %ld/10 completed.", chrNo + 1, l + 1 );
			// fprintf ( stderr, "*" );
			l++;
		}
	}
	//	Obtain the max number of overlaps.
	for ( i = chrPosition[ chrNo ]; i < chrPosition[ chrNo + 1 ]; ++i ) {
		sumSNP[ i ].ptInvolved[ tgPool ]  = sumSNP[ i ].sum[ tgPool ] = 0;																//	sumSNP[].sum is cleared for all SNP sites.
	}
	for( i2 = 0; i2 < noRegions; ++ i2 ) {
		if( ( chrPosition[ chrNo ] <= includeRegion[ i2 ].start ) && ( includeRegion[ i2 ].end <= chrPosition[ chrNo + 1 ] - 1 ) ) {
			for ( i = includeRegion[ i2 ].start; i <= includeRegion[ i2 ].end; ++i ) {						//	j moves within a region.
				sumSNP[ i ].ptInvolved[ tgPool ] = sumSNP[ i ].sum[ tgPool ] = sumSNP[ i ].foldOverlap[ noSubjects ];				//	Here, sumSNP[].sum means the number of patients sharing RCHHs with the representative.
				if ( maxOverlapChr[ chrNo ] < sumSNP[ i ].sum[ tgPool ] ) maxOverlapChr[ chrNo ] = ( int )sumSNP[ i ].sum[ tgPool ];		//	sumSNP[i].sum harbors the number of patients who share a common fragment.
			}
		}
	}
}
