//
//	readAnnotation.c
//
//	Programming:	Koich Hagiwara
//

#include	"myHeader.h"

extern	pSummerizedSNPdata sumSNP;
extern	pSummerizedSNPdata *pASNP;
extern	pSNPdata SNP;
extern	pSNPdata* pDSNP;
extern	int		lastChromosome;
extern	long	*chrPosition;
extern	double	totalGenetDistance, totalGenetDistanceWOX, totalGenetDistanceWOXWOG;
extern	double	averageMajorAlleleFrequencyWOX, averageMinorAlleleFrequencyWOX;
extern	long	totalAnnotationDataCount;
extern	int	noTgPool[ 2 ];
extern	long	*nEffectiveSNP[ 2 ];
extern	double memoryAllocated;
extern	long	largeGap;

// Read the annotation.txt file and set the data into tSumSNP[].  
// SNPs with a correct chromosome number, a position number and an rsID are considered valid.
void readAnnotation( char *pathToAnnotationFile )	
{
	FILE	*annotationFile;
	char	stringBuffer1[ STRINGBUFFERSIZE ];
	char	flag;
	struct	summerizedSNPdata  tSumSNP;
	long	i, j, k, start, end;
	float	ftemp;
	//	totalAnnotationDataRead -- Total number of annotation data read.
	long	totalAnnotationDataRead;
	
	//	totalGenetDistanceLocal -- Total genetic distance of the chromosomes.
	//	totalGenetDistanceWOXLocal -- Total genetic distance of the autosomes.
	//	totalGenetDistanceWOXWOGLocal -- Total genetic distance of the autosomes excluding gaps.
	int		tempInt;
	int		noParts;
	char	**parts;
	
	if ( ( char ** ) NULL == ( parts = ( char ** ) calloc ( MAX_NUMBER_OF_PARTS, sizeof ( char * ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n"); exit ( 1 ); }
	
	//	Open annotation file.
	lastChromosome = 0;
	totalAnnotationDataRead = 0;
	annotationFile = fopen( pathToAnnotationFile, "r" );
	while( ( char * )NULL != MyFgets( stringBuffer1, STRINGBUFFERSIZE, annotationFile ) ) {
		noParts = split( stringBuffer1, parts, "\t", MAX_NUMBER_OF_PARTS );
		sscanf( parts[ 0 ], "%d", &tempInt );
		if( tempInt > lastChromosome ) {
			lastChromosome = tempInt;
		}
		++totalAnnotationDataRead;
	}
	fclose( annotationFile );
	
	//	Allocate memory.
	if( ( pSummerizedSNPdata ) NULL == ( sumSNP = ( pSummerizedSNPdata ) calloc( totalAnnotationDataRead, sizeof( struct summerizedSNPdata ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	if( ( pSummerizedSNPdata *) NULL == ( pASNP = ( pSummerizedSNPdata * ) calloc( totalAnnotationDataRead, sizeof( pSummerizedSNPdata ) ) ) )  { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	for ( i = 0; i < totalAnnotationDataRead; ++i ) {
		if ( ( unsigned char * ) NULL == ( ( sumSNP + i )->gType[ 0 ] = ( unsigned char * ) calloc ( noTgPool[ 0 ] / 4 + 1, sizeof ( unsigned char ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n"); exit ( 1 ); }
		if ( ( unsigned char * ) NULL == ( ( sumSNP + i )->gType[ 1 ] = ( unsigned char * ) calloc ( noTgPool[ 1 ] / 4 + 1, sizeof ( unsigned char ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n"); exit ( 1 ); }
	}
	if( ( long * ) NULL == ( nEffectiveSNP[ 0 ] = ( long * ) calloc( ( noTgPool[ 0 ] + 1 ) * ( lastChromosome - EXCLUDE_X ), sizeof( long ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	if( ( long * ) NULL == ( nEffectiveSNP[ 1 ] = ( long * ) calloc( ( noTgPool[ 1 ] + 1 ) * ( lastChromosome - EXCLUDE_X ), sizeof( long ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	if( ( long * ) NULL == ( chrPosition = ( long * ) calloc( lastChromosome + 1, sizeof( long ) ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n" ); exit ( 1 ); }
	
	totalAnnotationDataRead = totalAnnotationDataCount = 0;
	annotationFile = fopen( pathToAnnotationFile, "r" );
	fprintf( stderr, "\nAnnotation data read = " );
	while( (char *)NULL != MyFgets( stringBuffer1, STRINGBUFFERSIZE, annotationFile) ) {
		noParts = split( stringBuffer1, parts, "\t", MAX_NUMBER_OF_PARTS );
		sscanf( parts[ 1 ], "%ld", &( tSumSNP.position ) );
		sscanf( parts[ 2 ], "%f", &( tSumSNP.AFrequency ) );
		if ( ( char * ) NULL == ( sumSNP[ totalAnnotationDataCount ].rsID = ( char * ) malloc ( strlen( parts[ 3 ] ) + 1 ) ) ) { fprintf( stderr, "ERROR: Failed to allocate memory\n"); exit ( 1 ); }
		strcpy( sumSNP[ totalAnnotationDataCount ].rsID, parts[ 3 ] );
		sscanf( parts[ 4 ], "%lf", &( tSumSNP.genetMap ) );
		flag = 0;			// If flag remains as 0, the data is valid.
		if ( totalAnnotationDataCount % 10000 == 0 ) fprintf( stderr, "%ld  ",totalAnnotationDataCount );	// showing progression of the program.
		
		if ( parts[ 0 ][ 0 ] == 'Y' || parts[ 0 ][ 0 ] == 'y' ) { continue; }
		if( parts[ 0 ][ 0 ] == 'X' || parts[ 0 ][ 0 ] == 'x' ) tSumSNP.chromosome = lastChromosome;
		else {
			sscanf( parts[ 0 ], "%d", &( tSumSNP.chromosome ) );
			if( tSumSNP.chromosome <= 0 || tSumSNP.chromosome > lastChromosome ) {
				flag = 1;											 //	If chromosome number is more than a 3-digit figure, the data is invalid.
				fprintf( stderr, "ERROR: Chromosome number error\n" );
			}
		}
		if( tSumSNP.position < 1 ) {
			flag = 1;					// If chromosome position is less than 1, it is invalid.
			fprintf( stderr, "ERROR: Position number error" );
		}
		if( ( tSumSNP.AFrequency < 0 ) || ( tSumSNP.AFrequency > 1 ) ){
			flag = 1;				// If AFrequency is less than 0, it is invalid.
			fprintf( stderr, "ERROR: Frequency value error" );
		}
		
		if( tSumSNP.genetMap <= 0 ) {
			flag = 1 ;				// If AFrequency is less than 0, it is invalid.
		} else if ( totalAnnotationDataCount != 0 ) {
			if ( ( sumSNP[ totalAnnotationDataCount - 1 ].chromosome == tSumSNP.chromosome ) && ( sumSNP[ totalAnnotationDataCount - 1 ].position < tSumSNP.position ) && ( sumSNP[ totalAnnotationDataCount - 1 ].genetMap > tSumSNP.genetMap ) ) {
				flag = 1;
				fprintf ( stderr, "%lf\t%lf\n", sumSNP[ totalAnnotationDataCount - 1 ].genetMap,  tSumSNP.genetMap );
			}
		}
		if ( flag == 1 ) fprintf( stderr, "ERROR: Genetic Map value error  " );
		++totalAnnotationDataRead;						// One data is read.
		
		// If flag==0, valid data is set in tSumSNP.
		
		if ( flag == 0 ) {										// A valid data came in.
			sumSNP[ totalAnnotationDataCount ].chromosome = tSumSNP.chromosome;	// Store the data and increment total valid data counter
			sumSNP[ totalAnnotationDataCount ].position = tSumSNP.position;	// Store the data and increment total valid data counter
			sumSNP[ totalAnnotationDataCount ].AFrequency = tSumSNP.AFrequency;	// Store the data and increment total valid data counter
			sumSNP[ totalAnnotationDataCount ].genetMap = tSumSNP.genetMap;	// Store the data and increment total valid data counter
			sumSNP[ totalAnnotationDataCount ].recordNo = totalAnnotationDataCount;	// Store the record number.
			++totalAnnotationDataCount;
		}
	}		// end of while
	
	fprintf( stderr, "\nNumber of SNPs read\t\t%ld\n", totalAnnotationDataRead );	// Total number of datapoints read.
	fprintf( stderr, "  Valid SNPs\t\t%ld\n\n", totalAnnotationDataCount );		// Total number of datapoints read.
	
	fclose( annotationFile );


	//	initialize pointer
	for ( i = 0; i < totalAnnotationDataCount; ++i ) {
		pASNP[ i ] = &( sumSNP[ i ] );
	}
	
	//	sort using rsID as a key.
	qsort( pASNP, totalAnnotationDataCount, sizeof( pSummerizedSNPdata ), compareAnnotRsID );
	
	for ( j = 0; j < lastChromosome; ++j ) {	// The position of the first SNP of chromosome 1 is stored in chrPosition[ 0 ].
		chrPosition[ j ] = 0;
	}
	for ( j = 0; j < lastChromosome; ++j ) {	// The position of the first SNP of chromosome 1 is stored in chrPosition[ 0 ].
		for ( i = 0; i < totalAnnotationDataCount; ++i ){
			if ( sumSNP[ i ].chromosome == ( j + 1 ) ) {
				chrPosition[ j ] = i;
				while( ( i < totalAnnotationDataCount ) && ( sumSNP[ i ].chromosome == ( j + 1 ) )  ) {
					i++;
				}
				for( k = j + 1; k <= lastChromosome; ++k ) {
					chrPosition[ k ] = i;
				}
				break;
			}
		}
	}
	chrPosition[ lastChromosome ] = totalAnnotationDataCount;	
	
	totalGenetDistance = 0.0;
	
	//	totalGenetDistanceLocal -- Total genetic distance of the chromosomes
	for ( j = 0; j < lastChromosome; ++j ) {
		if( chrPosition[ j + 1 ] != chrPosition[ j ] ) {
			totalGenetDistance += sumSNP[ chrPosition[ j + 1 ] - 1 ].genetMap - sumSNP[ chrPosition[ j ] ].genetMap;
		}
	}
	
	//	totalGenetDistanceWOX -- Total genetic distance of the autosomes.
	//	totalGenetDistanceWOXWOG -- Total genetic distance of the autosomes excluding gaps.
	totalGenetDistanceWOX = totalGenetDistanceWOXWOG = 0.0;
	
	for ( j = 0; j < lastChromosome - EXCLUDE_X; ++j ) {	// The number for the first SNP in the chromosome 1 is stored in chrPositionLocal[ 0 ].
		if( chrPosition[ j + 1 ] != chrPosition[ j ] ) {
			totalGenetDistanceWOX += sumSNP[ chrPosition[ j + 1 ] - 1 ].genetMap - sumSNP[ chrPosition[ j ] ].genetMap;
			start = end = chrPosition[ j ];
			for ( i = chrPosition[ j ], flag = 0; i < chrPosition[ j + 1 ]; ++ i ) {
				if ( i == chrPosition[ j + 1 ] - 1 ) {
					end = i;
					flag = 1;
				} else if ( sumSNP[ i + 1 ].genetMap - sumSNP[ i ].genetMap < largeGap ) {
					continue;
				} else {
					end = i;
					flag = 1;
				}
				if ( flag == 1 ) {
					totalGenetDistanceWOXWOG += sumSNP[ end ].genetMap - sumSNP[ start ].genetMap;
					start = end = i + 1;
				}
			}
		}
	}
	
	if( lastChromosome != 1 || chrPosition[ lastChromosome - EXCLUDE_X ] != 0 ) {
		for ( i = 0, averageMajorAlleleFrequencyWOX = averageMinorAlleleFrequencyWOX = 0.0; i < chrPosition[ lastChromosome - EXCLUDE_X ]; ++i){
			if ( sumSNP[ i ].AFrequency >= ( 1.0 - sumSNP[ i ].AFrequency) ) {
				averageMajorAlleleFrequencyWOX = averageMajorAlleleFrequencyWOX + sumSNP[ i ].AFrequency;
				averageMinorAlleleFrequencyWOX = averageMinorAlleleFrequencyWOX + ( 1.0 - sumSNP[ i ].AFrequency );
			} else {
				averageMajorAlleleFrequencyWOX = averageMajorAlleleFrequencyWOX + ( 1.0 - sumSNP[ i ].AFrequency);
				averageMinorAlleleFrequencyWOX = averageMinorAlleleFrequencyWOX + sumSNP[ i ].AFrequency;
			}
		}
		averageMajorAlleleFrequencyWOX = averageMajorAlleleFrequencyWOX / chrPosition[ lastChromosome - EXCLUDE_X ];
		averageMinorAlleleFrequencyWOX = averageMinorAlleleFrequencyWOX / chrPosition[ lastChromosome - EXCLUDE_X ];
	}
	
	fprintf( stderr, "Statistics of SNPs in the annotation file.\n" );
	fprintf( stderr, "     chr\t        from\t          to\t No. of SNPs\t  ave.hetero\tgenetic distance\n" );
	for ( j = 0; j < lastChromosome; ++j ) {
		if( chrPosition[ j + 1 ] != chrPosition[ j ] ) {
			fprintf( stderr, "%8d\t%12ld\t%12ld\t%12ld\t", sumSNP[ chrPosition[ j ] ].chromosome, sumSNP[ chrPosition[ j ] ].position, sumSNP[ chrPosition[ j + 1 ] - 1 ].position, chrPosition[ j + 1 ] - chrPosition[ j ] );
			for ( ftemp = 1.0, i = chrPosition[ j ]; i < chrPosition[ j + 1 ]; ++i ){
				ftemp += 1 - ( sumSNP[ i ].AFrequency * sumSNP[ i ].AFrequency + ( 1 - sumSNP[ i ].AFrequency ) * ( 1 - sumSNP[ i ].AFrequency ) );
			}
			ftemp = ftemp / ( chrPosition[ j + 1 ] - chrPosition[ j ] );				// chromosome 1
			fprintf( stderr, "%12lf\t%12lf\n", ftemp, sumSNP[ chrPosition[ j + 1 ] - 1 ].genetMap - sumSNP[ chrPosition[ j ] ].genetMap );
		}
	}
	
	fprintf ( stderr, "\nTotal genetic distance = %f\n", totalGenetDistance );
	fprintf ( stderr, "Total genetic distance (excluding X chromosome) = %f\n", totalGenetDistanceWOX );
	if( lastChromosome == 1 || chrPosition[ lastChromosome - EXCLUDE_X ] == 0 ) {
		fprintf ( stderr, "ERROR: No autosomes present in this analysis.\nTerminating Program....\n" );
		exit( 1 );
	} else {
		fprintf ( stderr, "The average of the frequency of the major allelic type excluding chromosome X %f\n", averageMajorAlleleFrequencyWOX );
		fprintf ( stderr, "The average of the frequency of the minor allelic type excluding chromosome X %f\n\n", averageMinorAlleleFrequencyWOX);
	}
	
	memoryAllocated = ( double )totalAnnotationDataRead * ( double )sizeof( struct summerizedSNPdata );
	memoryAllocated += ( double )totalAnnotationDataRead * ( double )sizeof( pSummerizedSNPdata );
	memoryAllocated += ( double )totalAnnotationDataRead * ( noTgPool[ 0 ] + 1 ) * sizeof ( unsigned short int );
	memoryAllocated += ( double )totalAnnotationDataRead * ( noTgPool[ 1 ] + 1 ) * sizeof ( unsigned short int );
	memoryAllocated += ( double )totalAnnotationDataRead * ( noTgPool[ 0 ] / 4 + 1 ) * sizeof ( unsigned char );
	memoryAllocated += ( double )totalAnnotationDataRead * ( noTgPool[ 1 ] / 4 + 1 ) * sizeof ( unsigned char );
	memoryAllocated += ( double )totalAnnotationDataRead * sizeof ( unsigned char );
	memoryAllocated += ( double )totalAnnotationDataRead * sizeof ( unsigned char );
	memoryAllocated += ( double )( noTgPool[ 0 ] + 1 ) * sizeof ( long );
	memoryAllocated += ( double )( noTgPool[ 1 ] + 1 ) * sizeof ( long );
	memoryAllocated += ( double )( lastChromosome + 1 ) * sizeof ( long );
	memoryAllocated += ( double )( lastChromosome + 1 ) * sizeof ( long );
	fprintf( stderr, "%.2f MegaBytes were allocated for data storage.\n", memoryAllocated / 1000.0 / 1000.0 );
	fprintf( stderr, "%.2f MegaBytes will be allocated for HH calculation.\n\n", chrPosition[ 1 ] * ( ( noTgPool[ 0 ] > noTgPool[ 1 ] )? noTgPool[ 0 ]: noTgPool[ 1 ] ) * sizeof( unsigned short int ) / 1000.0 / 1000.0 );
	
	return;
}
