//
//	setExcludedRegion.c
//
//	Programming:	Koich Hagiwara
//

#include "myHeader.h"

extern	pSummerizedSNPdata sumSNP;
extern	struct	incReg includeRegion[];
extern	double	totalGenetDistance, totalGenetDistanceWOX;
extern	long	*chrPosition;
extern	int		lastChromosome;

long	setRegion ( char *exclFileName, char *tgtFileName, long annotationNo )
{

	struct	startend {
		int		chromosome;
		long	position;
		float	genetMap;
	} t1, t2;
	
	char	*flag;
	FILE	*inf;
	char	stringBuffer1[ STRINGBUFFERSIZE ], stringBuffer2[ STRINGBUFFERSIZE ];
	long	i, j, frontEnd;
	long	start, end;
	long	regionCo = 0;
	double	genetDistance = 0.0;
	int	exclTgt;	// 0 if both excluded area file and target area file are not set.  1 if excluded area file is set.  2 if target area file is set.
	
	if ( NULL == ( flag = ( char *)malloc( annotationNo ) ) ) {
		fprintf( stderr, "ERROR: Cannot allocate memory.\n" );
		exit( 1 );
	}
	if( exclFileName[ 0 ] != 0 ) {
		exclTgt = 1;
	} else if( tgtFileName[ 0 ] != 0 ) {
		exclTgt = 2;
	} else {
		exclTgt = 1;
	}
	
	if( exclFileName[ 0 ] == 0 && tgtFileName[ 0 ] == 0 ) {
		for ( i = 0; i < lastChromosome - EXCLUDE_X; ++ i ) {
			includeRegion[ i ].start = chrPosition[ i ];
			includeRegion[ i ].end = chrPosition[ i + 1 ] - 1;
		}
		return ( lastChromosome - EXCLUDE_X );
	}
	
	if( ( exclFileName[ 0 ] != 0 ) && ( ( inf = fopen( exclFileName, "r" ) ) == NULL ) ) {
		fprintf( stderr, "ERROR: The excluded area file is not present.\n" ); exit( 1 );
	}
	if( ( tgtFileName[ 0 ] != 0 ) && ( ( inf = fopen( tgtFileName, "r" ) ) == NULL ) ) {
		fprintf( stderr, "ERROR: The target area file is not present.\n" ); exit( 1 );
	}

	//	All chromosomal regions are included at the beginning.
	for	( i = 0; i < annotationNo; i++ ) {
		if( exclTgt == 1 ) {
			// excluded area file
			flag[ i ] = 1;
		} else {
			// target area file
			flag[ i ] = 0;
		}
	}
	
	for ( i = 0, frontEnd = 0;
			( ( char * )NULL != MyFgets( stringBuffer1, STRINGBUFFERSIZE, inf ) ) && 
			( ( char * )NULL != MyFgets( stringBuffer2, STRINGBUFFERSIZE, inf ) ); ++i ) {
		sscanf( stringBuffer1, "%d %ld %f", &( t1.chromosome ), &( t1.position ), &( t1.genetMap ) );
		sscanf( stringBuffer2, "%d %ld %f", &( t2.chromosome ), &( t2.position ), &( t2.genetMap ) );
		for ( j = frontEnd; j < annotationNo; ++j ) {
			if ( ( sumSNP[ j ].chromosome == t1.chromosome ) && ( sumSNP[ j ].position == t1.position ) ) {
				frontEnd = start = j;	
				break;
			}
		}
		//	Weird!  We cannot find the starting point.
		if ( j == annotationNo ) {
			if( exclTgt == 1 ) {
				fprintf( stderr, "ERROR: The excluded area file is not compatible.\n" ); exit( 1 );
			} else {
				fprintf( stderr, "ERROR: The target area file is not compatible.\n" ); exit( 1 );
			}
		}
		for ( j = frontEnd; j < annotationNo; ++j ) {
			if ( ( sumSNP[ j ].chromosome == t2.chromosome ) && ( sumSNP[ j ].position == t2.position ) ) {
				end = j; frontEnd = j + 1;
				break;
			}
		}
		//	Weird!  We cannot find the end point.
		if ( j == annotationNo ) {
			if( exclTgt == 1 ) {
				fprintf( stderr, "ERROR: The excluded area file is not compatible.\n" ); exit( 1 );
			} else {
				fprintf( stderr, "ERROR: The target area file is not compatible.\n" ); exit( 1 );
			}
		}
		for ( j = start; j <= end; ++j ) {
			if( exclTgt == 1 ) {
				// excluded area file
				flag [ j ] = 0;
			} else {
				// target area file
				flag [ j ] = 1;
			}
		}
	}
	for	( frontEnd = 0; frontEnd < annotationNo; ) {
		for ( j = frontEnd; j < annotationNo; ++j ) {
			if ( flag [ j ] == 1 ) {
				frontEnd = start = j;
				break;
			}
		}
		if ( j == annotationNo ) break;										//	If chromosome ends while we are outside.
		for ( j = frontEnd; j < annotationNo; ++j ) {						//	In a region.
			if ( flag [ j ] == 0 ) {										//	If a region ends.
				end = j - 1;
				frontEnd = j;
				break;
			}
			if ( j == annotationNo - 1 ) {									//	If all chromosome have been searched for.  We are still inside.
				end = j;
				frontEnd = j + 1;											//	To exit the loop.
				break;
			}
			if ( sumSNP[ j ].chromosome != sumSNP[ j + 1 ].chromosome ) {	//	Chromosome changed.
				end = j;
				frontEnd = j + 1;
				break;
			}
		}
		includeRegion[ regionCo ].start = start;
		includeRegion[ regionCo ++ ].end = end;
		genetDistance += sumSNP[ end ].genetMap - sumSNP[ start ].genetMap;
	}
	fprintf( stderr, "Total genetic distance included = %f.  Ratio is %f\n", genetDistance, genetDistance / ( ( EXCLUDE_X == 1 ) ? totalGenetDistanceWOX : totalGenetDistance ) );
	fclose( inf );
	free( flag );
	return ( regionCo );
}
